# -*- encoding: UTF-8 -*-
from __future__ import unicode_literals
import sys
import re
import os
import codecs

comment = re.compile(r"[\n#]")
ids = re.compile(r"\w+:\s*\*?\w+(,\s*\*?\w+)*")
langu = re.compile(r"\[.+=.+\]\s*")
titl = re.compile(r"\w+\s*=\s*")
helptexts = []

xdl_header = """<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE dlg:window PUBLIC "-//OpenOffice.org//DTD OfficeDocument 1.0//EN" "dialog.dtd">
<dlg:window xmlns:dlg="http://openoffice.org/2000/dialog" xmlns:script="http://openoffice.org/2000/script" dlg:id="%s" dlg:left="101" dlg:top="52" dlg:width="196" dlg:height="72" dlg:closeable="true" dlg:moveable="true" dlg:withtitlebar="false">
 <dlg:bulletinboard>
"""
xdl_footer = """</dlg:bulletinboard>
</dlg:window>
"""
xdl_group = '<dlg:fixedline dlg:id="%s" dlg:tab-index="%d" dlg:left="5" dlg:top="%d" dlg:width="240" dlg:height="10" dlg:value="&amp;%s"/>\n'
xdl_item = '<dlg:checkbox dlg:id="%s" dlg:tab-index="%d" dlg:left="%d" dlg:top="%d" dlg:width="%d" dlg:height="10" dlg:value="&amp;%s" dlg:checked="%s" %s/>\n'

xcs_header = """<?xml version="1.0" encoding="UTF-8"?>
<oor:component-schema xmlns:oor="http://openoffice.org/2001/registry" xmlns:xs="http://www.w3.org/2001/XMLSchema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
oor:name="%s" oor:package="org.openoffice" xml:lang="en-US">
<info>
<desc>Contains the options data used for the test extensions.</desc>
</info>
<templates>
"""
xcs_leaf_header = r"""
                <group oor:name="%s">
                        <info>
                                <desc>The data for one leaf.</desc>
                        </info>
"""
xcs_leaf = r"""<prop oor:name="%s" oor:type="xs:string">
                                <value></value>
</prop>
"""
xcs_leaf_footer = r"""                </group>
"""
xcs_component_header = """        </templates>
        <component>
                <group oor:name="Leaves">
"""
xcs_component = """
                        <node-ref oor:name="%s" oor:node-type="%s"/>
"""
xcs_footer = """                </group>
        </component>
        
</oor:component-schema>
"""
xcu_header = u"""<?xml version='1.0' encoding='UTF-8'?>
<!DOCTYPE oor:component-data SYSTEM "../../../../component-update.dtd">
<oor:component-data oor:name="OptionsDialog" oor:package="org.openoffice.Office" xmlns:oor="http://openoffice.org/2001/registry" xmlns:xs="http://www.w3.org/2001/XMLSchema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
        <node oor:name="Nodes">
                <node oor:name="org.openoffice.lightproof" oor:op="fuse">
                        <prop oor:name="Label">
                                <value xml:lang="en">Dictionaries</value>
                                <value xml:lang="hu">Szótárak</value>
                        </prop>
                        <node oor:name="Leaves">
"""
xcu_node_header = """
                                <node oor:name="org.openoffice.lightproof.%s" oor:op="fuse">
                                        
                                        <prop oor:name="Id">
                                                <value>org.openoffice.comp.pyuno.lightproof.oxt.%s</value>
                                        </prop>
                                        
                                        <prop oor:name="Label">
"""
xcu_node = """
                                                <value xml:lang="%s">%s</value>
"""
xcu_node_footer = """
                                        </prop>
                                        
                                        <prop oor:name="OptionsPage">
                                                <value>%%origin%%/%s.xdl</value>
                                        </prop>
                                        
                                        <prop oor:name="EventHandlerService">
                                                <value>org.openoffice.comp.pyuno.LightproofOptionsEventHandler.%s</value>
                                        </prop>
                                        
                                </node>
"""
xcu_footer = """
                        </node>
                </node>
        </node>
</oor:component-data>
"""

indexes = {}
indexes_def = {}
modules = {}

def create_xdl(pkg, lines, target, lang):
    global indexes
    global indexes_def
    global modules
    indexes[lang] = []
    indexes_def[lang] = []
    modules[lang] = []
    f2 = ""
    state = 0
    f2n = "dialog/" + lang + ".xdl"
    f2 = f2 + xdl_header%lang

    k = 0
    k2 = 0
    lin = 0
    ok = False
    for i in lines:
        i = i.strip()
        if "=" in i and r"\n" in i:
            helptexts.append(i.split("=")[0])
    for i in lines:
        i = unicode(i.strip().replace(r"\n", "@#@") + "\n", "UTF-8")
        lin = lin + 1
        if not comment.match(i):
            if state == 0:
                ok = True
                if ids.match(i.strip()):
                    j = i.strip().split(":")
                    f2 = f2 + xdl_group%(j[0].strip(), k, k2 * 10 + 5, j[0].strip())
                    for l in j[1].split(","):
                        k = k + 1
                        k2 = k2 + 1
                        l = l.strip()
                        la = l.split(" ")
                        l3 = 0
                        itemlen = int(240 / len(la)) 
                        for l2 in la:
                            if l2 != "-":
                                checked = "false"
                                if l2[0] == '*':
                                    checked = "true"
                                    l2 = l2[1:]
                                    indexes_def[lang] += [l2]
                                indexes[lang] += [l2]
                                helptext = ""
                                if l2 in helptexts:
                                    helptext = "dlg:help-text=\"&amp;hlp_" + l2 + "\""
                                f2 = f2 + xdl_item%(l2, k, 10 + itemlen * l3, k2 * 10 + 5, itemlen, l2, checked, helptext)
                                l3 = l3 + 1
                                k = k + 1
                    k2 = k2 + 1
                    k = k + 1
                else:
                    ok = False
            if langu.match(i.strip()):
                if "xdl" in f2n:
                    f2 = f2 + xdl_footer
                target.writestr(f2n, f2)
                f2 = ""
                i = i.strip()
                langname = i[1:i.find("=")]
                modules[lang] += [langname[:langname.find("_")], i[i.find("=")+1:-1]]
                f2n = "dialog/" + lang + "_" + langname + ".properties"
                state = state + 1
                if state == 1:
                    target.writestr("dialog/" + lang + "_" + langname + ".default", "")
            elif titl.match(i.strip()):
                hlp = i.encode("unicode-escape").replace(r"\n","\n").replace(r"\t","\t").replace(r"\x","\\u00").split("@#@", 1)
                if len(hlp) > 1:
                    helptexts.append(hlp[0].split("=")[0])
                    f2 = f2 + "hlp_" + hlp[0].split("=")[0] + "=" + hlp[1]
                    hlp[0] = hlp[0] + "\n"
                f2 = f2 + hlp[0]
            elif not ok:
                print ("Syntax error in line %d: %s" %(lin, i))
    if "xdl" in f2n:
        f2 = f2 + xdl_footer
    target.writestr(f2n, f2)

#def c(pkg, author, language, inpdir, target, prgtarget, dlg):

def c(pkg, dlgdata, target, lang):

 # create xdl dialog data files
 create_xdl(pkg, dlgdata, target, lang)

 s = xcs_header% ("Lightproof_" + pkg)
 for i in indexes:
    s = s + xcs_leaf_header%i + \
        xcs_leaf*len(indexes[i])%tuple(indexes[i]) + \
        xcs_leaf_footer
 s = s + xcs_component_header
 for i in indexes:
    s = s + xcs_component%(i,i)

 target.writestr("dialog/OptionsDialog.xcs", s + xcs_footer)

 s = "" 
 for i in indexes:
    s = s + xcu_node_header%(pkg, pkg) + \
        xcu_node*(len(modules[i])/2)%tuple(modules[i]) + \
        xcu_node_footer%(i, pkg)
 target.writestr("dialog/OptionsDialog.xcu", (xcu_header + s + xcu_footer).encode("utf-8"))

 # python resource file

 s = """lopts = {}
lopts_default = {}
"""
 for i in indexes:
    s = s + "lopts['" + i + "'] = " + str(indexes[i]) + "\n"
 for i in indexes:
    s = s + "lopts_default['" + i + "'] = " + str(indexes_def[i]) + "\n"
 target.writestr("pythonpath/lightproof_opts_%s.py"%pkg, s)






# -*- encoding: UTF-8 -*-
#from __future__ import unicode_literals
import sys, os, zipfile, traceback, Dialog
try:
    import ConfigParser as cp    # configparser in Python 3
    uni = lambda i: unicode(i, "UTF-8")
except:
    import configparser as cp
    uni = str

import pythonpath.lightproof_compile___implname__
from string import Template

def dist(fn, a):
    a['locales'] = a["locales"].replace("_", "-")
    a['loc'] = str(dict([[i, [i[0:2], i[3:5], ""]] for i in a["locales"].split(" ")]))
    distname = a['implname'] + "-" + a['version'] + '.oxt'
    z = zipfile.ZipFile(distname, mode='w', compression = zipfile.ZIP_DEFLATED)
    f = open(fn + ".dat",'r')
    code = pythonpath.lightproof_compile___implname__.c(uni(f.read()), a['lang'])
    a["code"] = code["code"]
    a['data'] = code["rules"]

    for i in ["META-INF/manifest.xml", "description.xml", "Linguistic.xcu", "Lightproof.py", \
        "pythonpath/lightproof_handler___implname__.py", "pythonpath/lightproof_impl___implname__.py", \
        "pythonpath/lightproof___implname__.py" ]:
        z.writestr(i.replace("__implname__", a["implname"]), Template(uni(open(i, "r").read())).safe_substitute(a))

    for i in a["extras"].split(","):
        z.writestr(i.strip().replace("../", "").replace("__implname__", a["implname"]), \
        open(fn[:fn.rfind("/")+1] + i.strip()).read())

    try:
        d = open(fn + ".dlg", "r").readlines()
        Dialog.c(a["implname"], d, z, a["lang"])
    except:
        z.writestr("pythonpath/lightproof_opts_%s.py"%a["implname"], "")

if len(sys.argv) == 1:
    print ("""Synopsis: python make.py config_file
eg. python make.py src/en/en.cfg""")
    sys.exit(0)

fArgs = cp.SafeConfigParser()
for i in sys.argv[1:]:
    try:
        fArgs.read(i)
        dist(i[:-4], fArgs._sections['args'])
    except:
        print (traceback.format_exc())
        print ("missing config file or options: " + str(i))
        sys.exit(0)



from __future__ import unicode_literals
import sys, re, traceback

repl = {}
tests = []
comp = []
modes = ["[Word]", "[word]", "[Char]", "[char]", "[code]"]
mode = "[Word]"
code = u""
language = ""
oldlinenums = {}

def prepare_for_eval(s):
    s = re.sub("(affix|spell|morph|stem|option|suggest|generate)\(", r'\1(LOCALE,', s)
    s = re.sub(r"word\(\s*(\d)", r'word(s[m.end():],\1', s)       # word(n)
    s = re.sub(r"word\(\s*-(\d)", r'wordmin(s[:m.start()],\1', s) # word(-n)
    s = re.sub(r"[\\](\d)", r'm.group(\1)', s)
    s = re.sub("[{]([^}]+)}", r'm.group("\1_1")', s)
    return s


def mysplit(s, line, oldline, debug):
    global repl
    global tests
    global comp
    global modes
    global mode
    orig = s
    if s[0:1] == '[':
        if s.strip() in modes:
            mode = s.strip()
            return None
        elif re.match(r"\[\w+\]$", s.strip()):
            #sys.stderr.write("Unknown mode: " + s + "\n")
            return oldline
    dec = 0
    exprep = 0 # replacement is a Python expression (beginning with sign =)
    condition = False
    ngroup = 0 # back reference group number that will be used for error positioning
    # description
    c = re.search("\s#\s", s)
    com = u""
    if c:
        try:
            c = c.start(c.lastindex - 1)
        except:
            c = c.start()
        com = s[c+2:].strip()
        if com[0:1] == "=":
            com = prepare_for_eval(com)
        s = s[:c]
    m1 = re.search("<-", s)
    m2 = re.search("-\d*>", s)
    if m1 and m2:
        condition = prepare_for_eval(s[m1.end(0): m2.start(0)].strip())
        s = s[0:m1.start(0)] + s[m2.start(0):]
    if s[0:1] == '"':
        # quoted
        pos = s[1:].find('"')
        while pos > 0 and s[pos] == '\\':
            pos = s[pos:].find('"')
        s1 = s[1:pos+1]
        s2 = s[pos+2:].strip()
    else:
        m = re.compile("-\d*>").search(s)
        if not m:
            m = re.compile("[_a-zA-Z][_a-zA-Z0-9]*").match(s)
            if not m:
                # syntax error
                return oldline
            s1 = m.group(0)
            s2 = s[m.end(0):].strip()
            # replace previous definitions
            for i in repl:
                ire = re.compile("[{]" + i + "}")
                if re.search(ire, s2):
                    s2 = ire.sub(repl[i], s2)
            # make named group
            s2 = "(?P<" + m.group(0) + ">" + s2 + ")"
            dec = 1
        else:
            s1 = s[0:m.start(0)].strip()
            if re.match("TEST: ", s1): # deprecated
                tests += [[s1[5:].strip(), s[m.end(0):].strip(), oldline]]
                return None
            s2 = s[m.start(0):].strip()
    m = re.compile("-(\d*)>").match(s2)
    if dec!= 1 and m:
        s2 = s2[m.end(0):].strip()
        if m.group(1):
            ngroup = int(m.group(1))
    elif dec!=1:
        # syntax error
        return oldline
    if s2[0:1] == '=':
        exprep = 1
    if s2[0:1] == '"' and s2[-1:]=='"':
        s2 = s2[1:-1]
    if dec==1:
        repl[s1] = s2
        return None
    else:
        for i in repl:
            s1 = re.sub("[{]" + i + "}", repl[i], s1)
        # modes
        if mode == "[Word]" or mode == "[word]":
            if s1[0] == '^':
                s1 = r"((?<=[!?.] )|^)" + s1[1:] + r"(?![-\w\u2013\u00AD])"
            else:
                s1 = r"(?<![-\w\u2013.,\u00AD])" + s1 + r"(?![-\w\u2013\u00AD])"
        # modes with casing ([Word] or [Char])
        if mode == "[Word]" or mode == "[Char]":
            s1 = "(?i)" + s1
        # add Unicode flag to the regex by (?u)
        s1 = re.sub("[(][?][iI][)]", "(?iu)", s1)
        if not re.match("[(][?][iI][uU][)]", s1):
            s1 = "(?u)" + s1
        else:
            # casing
            lu = re.compile("(?u)\w")
            s3 = u""
            state = 0
            for i in range(0, len(s1)):
                c = s1[i]
                if c == "[":
                    state = 1
                if state == 1 and c == "]":
                    state = 0
                if c == "<" and i > 3 and s1[i-3:i]=="(?P":
                    state = 2
                if state == 2 and c == ">":
                    state = 0
                if c == "?" and i > 0 and s1[i-1:i]=="(":
                    state = 5
                if state == 5 and c == ")":
                    state = 0
                if lu.match(c) and c.islower() and state == 0:
                    if c=="i" and (language == "tr" or language == "az"):
                        s3 += u"[\u0130" + c + "]"
                    else:
                        s3 += "[" + c.upper() + c + "]"
                elif lu.match(c) and c.islower() and state == 1 and s1[i+1:i+2] != "-":
                    if s1[i-1:i] == "-" and s1[i-2:i-1].islower():  # [a-z] -> [a-zA-Z]
                        s3 += c + s1[i-2:i-1].upper() + "-" + c.upper()
                    elif c=="i" and (language == "tr" or language == "az"):
                        s3 += u"\u0130" + c
                    else:
                        s3 += c.upper() + c
                else:
                    s3 += c
                if c == "\\":
                    state = 4
                elif state == 4:
                    state = 0
            s1 = s3
        s1 = renum("[?]P<([^<_]*)>", s1, "?P<")
        if exprep == 0:
            s2 = re.sub("[{]([_a-zA-Z][_a-zA-Z0-9]*)}", r"\\g<\1>", s2)
            s2 = renum(r"\\g<([^<_]*)>", s2, r"\\g<")
        else:
            s2 = prepare_for_eval(s2)
    # check
    if re.compile("[(][?]iu[)]").match(s1):
        cap = True
        sc = re.sub("[(][?]iu[)]", "(?u)", s1)
    else:
        cap = False
        sc = s1
    try:
        compr = re.compile(sc)
        if not condition:
            comp += [[compr, s2, com, cap, line]]
    except Exception as e:
        raise Exception(str(e), oldline)
    if debug:
        return [s1, s2, com, condition, ngroup, oldline]
    return [s1, s2, com, condition, ngroup]

# group renum (<groupname> -> <groupname_1> etc.)
def renum(regex, s1, beg):
    j={}
    mr = re.compile(regex)
    m = mr.search(s1)
    nl = s1.find("\\n")
    while m:
        # restart numbering in new lines
        if nl > -1 and m.start(0) > (nl + 1):
            j={}
            nl = s1[m.start(0):].find("\\n")
            if nl > -1:
                nl = m.start(0) + nl
        n = m.group(1)
        if n in j:
            j[n] += 1
        else:
            j[n] = 1
        s1 = re.sub(mr, beg + n + "_" + str(j[n]) + ">", s1, 1)
        m = mr.search(s1)
    return s1

def cap(a, iscap):
    global language
    if iscap:
        for i in range(0, len(a)):
            if a[i][0:1] == "i":
                if language == "tr" or language == "az":
                    a[i] = u"\u0130" + a[i][1:]
                elif a[i][1:2] == "j" and language == "nl":
                    a[i] = "IJ" + a[i][2:]
                else:
                    a[i] = "I" + a[i][1:]
            else:
                a[i] = a[i].capitalize()
    return a

def c(rules, lang, debug = False):
    global language
    global code
    global oldlinenums
    language = lang
    r = re.compile("[\n#]")
    code = ""
    dic = []
    oldlinenums = {}

    lines = rules.split("\n")
    lines2 = []
    result = {}


    cm = 0
    lnums = 1
    for i in lines:
        if i.strip() in modes:
            if i.strip() == "[code]":
                cm = 1
                continue
            else:
                cm = 0
        if cm == 1:
            code = code + i + "\n"
        elif len(i.strip()) > 0:
            lines2 = lines2 + [i]
            oldlinenums[i] = lnums
        lnums = lnums + 1

    lines = lines2

    # concatenate multiline commands
    # last tabulator + comment is the message
    l = u""
    comment = 0
    for i in range(len(lines)-1, -1, -1):
        if re.match("\t", lines[i]):
            if not (comment and re.match("\t+#", lines[i])):
                l = lines[i].strip() + " " + l
            if re.search("#", lines[i]):
                comment = 1
            del lines[i]
        elif l != "":
            lnums = oldlinenums[lines[i]]
            lines[i] = lines[i].strip() + " " + l
            oldlinenums[lines[i]] = lnums
            l = ""
            comment = 0

    # processing
    for i in range(0, len(lines)):
        if not r.match(lines[i]):
            item = mysplit(lines[i].strip(), i + 1, oldlinenums[lines[i]], debug)
            if item != None:
                if type(item) == type(1):
                    raise Exception("Syntax error in line ", item)
                dic = dic + [item]

    result["rules"] = dic
    code = re.sub(r"(?<![\\])\\n", "\n", code.encode("unicode-escape").decode('ascii'))

    result["code"] = re.sub(r"(?<![\\])\\t", "\t", code).replace(r"\\n", r"\n").replace(r"\\t", r"\t")

    return result



# -*- coding: UTF-8 -*-
import lxml.etree as ET
import sys
import os
import codecs
import copy

did_you_mean = False
number = 0
amount = 0
to_implement = 0
no_support = 0


class RuleModel(object):
    '''parse() method returns rule'''
    def __init__(self, rule):
        self.rule = rule
        self.pattern = rule.find('pattern')
        self.pattern_descendants = RuleModel.element_descendants(self.pattern)
        self.message = rule.find('message')
        self.message_descendants = RuleModel.element_descendants(self.message)
        self.items = RuleModel.nesting_list_killer(
                    [i.items() for i in self.message.iterdescendants()]
                    + [i.items() for i in self.pattern.iterdescendants()])
        self.keys = RuleModel.nesting_list_killer(
                    [i.keys() for i in self.message.iterdescendants()]
                    + [i.keys() for i in self.pattern.iterdescendants()])

    @staticmethod
    def nesting_list_killer(lst):
        result = []
        for element in lst:
            if isinstance(element, (str, unicode, tuple)):
                result.append(element)
            else:
                result += RuleModel.nesting_list_killer(element)
        return set(result)

    @staticmethod
    def key_remover(iterable, keys_list):
        keys = copy.copy(keys_list)
        it = list(copy.copy(iterable))
        for key in keys:
            try:
                it.remove(key)
            except ValueError:
                pass
        return it

    @staticmethod
    def space_killer(old_result):
        '''space killer, char dealer'''
        result = [i.strip() for i in old_result[1:]]

        for item in result:
            if (item in ('.', ',', ':', '!', '?', "'", '"', '([.])')
                and item in result
                or item.startswith('\\n')
                ):
                indx = result.index(item)
                if indx != 0:
                    result[indx - 1] += item
                    result.remove(item)
                else:
                    result = [(item + result[:][1]), ] + result[2:]

        for item in result:
            if '\\n' in item:
                result[result.index(item)] = '\\n'.join(i.strip()
                                                    for i in item.split('\\n'))

            if item.startswith('([.])') or item.endswith('([.]))'):
                indx = result.index(item)
                if indx != 0:
                    result[indx - 1] += item
                    result.remove(item)

        for item in result:
            if item.endswith('([.]))'):
                indx = result.index(item)
                if indx != 0:
                    result[indx - 1] += item
                    result.remove(item)
#            if '|' in item:
#                result[result.index(item)] = '\\n'.join(i.strip()
#                                                    for i in item.split('|'))

        for item in result:
            if item.endswith("'") and (len(result) - result.index(item) >= 2):
                indx = result.index(item)
                result[indx] += result[indx + 1]
                result.remove(result[indx + 1])

        result.insert(0, old_result[0])
        return result

    @staticmethod
    def element_descendants(element):
        return ((i, i.tag, tuple(i.keys()), i.getchildren(), i.text, i.tail)
                                            for i in element.iterdescendants())

    def parse(self, attr):
        result = []
        suggestions = []

        if attr == 'pattern':
            target = self.pattern_descendants
        elif attr == 'message':
            target = self.message_descendants
        else:
            raise AttributeError("wrong attr for parse()'s target ")

        for (element, tag, keys, _children, text,
                                            tail) in target:
# start for #
            '''unsupported pattern keys'''
            if (self.keys
                and 'postag' in self.keys
                or 'postag_regexp' in self.keys
                or 'negate_pos' in self.keys
                ):
                if ('postag', 'SENT_START') in self.items:
                    r = copy.copy(self.items)
                    r.remove(('postag', 'SENT_START'))
                    r = [i[0] for i in RuleModel.nesting_list_killer(r)]
                    if ('postag' in r
                        or 'postag_regexp' in r
                        or 'negate_pos' in r):
                        result = [[], 'Not supported', ]
                        return result
                    else:
                        pass
                else:
                    result = [[], 'Not supported', ]
                    return result

            '''not implemented features'''
            if (self.keys
                and 'regexp_match' in self.keys
                or 'regexp_replace' in self.keys
                or 'regexp_replace' in self.keys
                or 'case_conversion' in self.keys
#                or tag == 'match'
#                or 'skip' in self.keys
#                or 'negate' in keys
#                or 'spacebefore' in keys
#                or 'inflected' in keys
#                or tag == 'exception'
                or ('postag', 'SENT_START') in self.items
                ):
                if not 'Not supported' in result:
                    result = [[], 'Not implemented', ]
                    return result

            if not len(result):
                result.append(element.items())
            else:
                result[0] += element.items()

            #exception tag with scope attr rule
            if (tag == 'exception' and 'scope' in keys
                and not RuleModel.key_remover(keys, ['scope', ])):
                ''' exception tag with only the scope key'''
                if element.get('scope') == 'previous':
                    result.append('(?<!%s)' % text)
                if element.get('scope') == 'next':
                    result.append('(?!%s)' % text)
                if element.get('scope') == "current":
                    result.append('(?=%s)' % text)
                if tail is not None:
                    result[-1] += ' (%s)' % tail

            # regexp attr rule
            if ('regexp' in keys
                and not RuleModel.key_remover(keys, ['regexp',
                                                     'inflected',
                                                     'skip'
                                                     ])
                and not element.getparent() == 'marker'):
                '''elements with only the regexp attr'''
                if text and text.strip():
                    result.append('(%s)' % text.strip())

            # token tag rule
            if tag == 'token':
                if not RuleModel.key_remover(keys, ['inflected',
                                                    'skip'
                                                    ]):
                    '''common tokens with no attrs'''
                    if text and text.strip():
                        if text is not '.':
                            result.append('(%s)' % text.rstrip())
                        else:
                            result.append('([.])')
                    else:
                        if not element.getchildren():
                            result.append('()')

            # marked tag rules
            if (element.tag == 'marker' and not element.getchildren()):
                '''markers without children -> error'''
                raise TypeError("'markered' elements in %s"
                                    % self.rule.get('id'))
            if element.getparent().tag == 'marker':
                children_list = element.getparent().getchildren()
                if len(children_list) == 1:
                    pass
                    #result[-1] = '(' + result[-1] + ')'
                else:
                    if children_list[0] == element:
                        if "regexp" not in element.keys():
                            if result[-1][0] != '(':
                                result[-1] = '((' + result[-1] + ')'
                            else:
                                result[-1] = '(' + result[-1]
                    elif len(result) > 2:
                        if (element in children_list[1:-1]
                              and "regexp" not in element.keys()):
                            if not result[-1][0] == '(':
                                result[-1] = '(' + result[-1] + ')'

                        elif children_list[-1] == element:
                            """ ')' after last element"""
                            if result[-1][0] != '(':
                                result[-1] = '(' + result[-1] + '))'
                            elif result[-1][0] == '(':
                                result[-1] += ')'

            #negate attr rule
            if (('negate', 'yes') in element.items()
                and not RuleModel.key_remover(keys, ['regexp',
                                                     'default',
                                                     'skip',
                                                     'negate'])):
                '''negate="yes" attribute'''
                if not RuleModel.key_remover([i for i in text],
                                        ('.', ',', ':', '!', '?', "'", '"',)):
                    result.append('([^%s])' % text)
                elif 'regexp' in keys:
                    result.append('(^(%s))' % text)
                else:
                    result.append('(^%s)' % text)

            #spacebefore attr rule
            if ('spacebefore' in keys
                and not RuleModel.key_remover(keys, ['default',
                                                     'spacebefore',
                                                     'skip',
                                                     'inflected'])):
                if text and text.strip():
                    if not self.pattern.getchildren()[0] == element:
                        if element.get('spacebefore') == 'no':
                            result[-1] += text.strip()
                        else:
                            result[-1] = result[-1] + ' ' + text.strip()
                    else:
                        if element.get('spacebefore') == 'no':
                            result.append('(w*)%s' % text.strip())
                        else:
                            result.append('(w*) %s' % text.strip())
            #spacebefore + regexp attr rule
            if ('spacebefore' in keys
                and 'regexp' in keys
                and not RuleModel.key_remover(keys, ['default',
                                                     'spacebefore',
                                                     'regexp',
                                                     'skip',
                                                     'inflected'])):
                if text and text.strip():
                    if not self.pattern.getchildren()[0] == element:
                        if element.get('spacebefore') == 'no':
                            result[-1] = ('(%s%s)'
                                          % (result[-1].translate(None, '()'),
                                             text.strip()))
                        else:
                            result[-1] = ('(%s %s)'
                                          % (result[-1].translate(None, '()'),
                                             text.strip()))
                    else:
                        if element.get('spacebefore') == 'no':
                            result.append('()%s' % text.strip())
                        else:
                            result.append('(() %s)' % text.strip())
            # skip attr rule
            if 'skip' in keys:
                if element.get('skip') == '-1':
                    result.append('(\w+ )*')
                else:
                    result.append('(\w+ ){0,%d}' % int(element.get('skip')))

            # rules for message
            # suggestion tag with no attrs or suggestion match tag with 1 attr
            if tag == 'suggestion' and not RuleModel.key_remover(keys, 'no'):
                if text and text.strip() and not keys:
                    '''message suggestions appearance'''
                    if [item.startswith('=') for item in result[1:]]:
                        result.append('\\n%s' % text.strip())
                        suggestions.append('%s' % text.strip())
                    else:
                        result.append('%s' % text.strip())
                        suggestions.append('%s' % text.strip())

                # suggestion match rules
                for el in element.getchildren():
                    if el.tag == "match":
                        match_no = int(el.get('no'))
#                       match_el = self.pattern.getchildren()[match_no- 1]
#                       print 'ZZZ', match_el.text
                        result.append('\\%s' % (match_no))
                        suggestions.append('\\%s' % (match_no))
                        if el.tail and el.tail.strip():
                            result.append('%s' % el.tail.strip())
                            suggestions.append('%s' % el.tail.strip())
                if tail and tail.strip():
                    suggestions.append(tail.strip())

# end for #
        if attr == 'message':
            if self.message.getchildren() and not result:
                '''in case of empty message'''
                raise TypeError('empty message!')

            '''message verbosity'''
            global did_you_mean
            if did_you_mean:
                result.append('#Did you mean?')
            else:
                result.append('#%s' % self.message.text.strip())
                if self.message.tail and self.message.tail.strip():
                    result.append('%s' % self.message.tail.strip())

                if suggestions is not None:
                    for suggestion in suggestions:
                        if suggestion is not None:
                            result.append('%s' % suggestion)

        result = RuleModel.space_killer(result)
        return result


class RuleView(object):
    '''adaptes rule's API, printer() method prints rule'''
    def __init__(self, rule, category):
        self.parse_pattern = RuleModel(rule).parse('pattern')
        self.pref = self.parse_pattern[0]
        self.pattern = self.parse_pattern[1:]
        self.parse_message = RuleModel(rule).parse('message')
        self.ix = self.parse_message[0]
        self.message = self.parse_message[1:]
        self.prefix = list(self.pref) + list(self.ix)
        self.id = rule.get('id')[0:15] if rule.get('id') else 'Unknown'
        self.rule = rule
        self.category = category
        self.supported_keys = ['regexp', 'scope', 'spacebefore', 'negate']
        self.key_to_print = None
        self.keys = RuleModel.nesting_list_killer([i.keys()
                                        for i in self.rule.iterdescendants()])

    @staticmethod
    def aggregate(iter1, iter2):
        set1 = set(iter1)
        set2 = set(iter2)
        aggregate = set1 & set2
        return list(aggregate)

    def sys_argv(self):
        supported_args = {'-short': 'Print \"Did you mean?\". '
                                        'Default: Print full message',
                          '-file': 'Dump rules to \'rules.txt\' file',
                          '-attr': 'Print <rule> tags\'s attr list '
                          'before rule if any',
                          '-show=attr': 'Incompatible with other '
                          ' keys except maybe -short :)'}
        if len(sys.argv) > 1:
            if '-short' in sys.argv:
                global did_you_mean
                did_you_mean = True
            if '-file' in sys.argv:
                sys.stdout = codecs.open('rules.txt',
                                         encoding='utf-8', mode='a+')
            if '-attr' in sys.argv:
                if self.prefix:
                    print sorted(list(set(self.prefix))), '=>',

            show_attr = [i for i in sys.argv if i.startswith('-show=')]
            if show_attr:
                if len(show_attr) == 1:
                    self.key_to_print = show_attr[0][6:]
                else:
                    raise TypeError('Too many "show" arguments')

            if False in [el in supported_args.keys() for el in sys.argv[1:]
                                            if not el.startswith('-show=')]:
                print '\n-*-sys.argv Error-*-'
                print '\nYou printed:  python',
                for i in sys.argv:
                    print i,
                print ('\nUsage: python %s [ Keys ]\n\n Keys:'
                       % sys.argv[0])
                for i, j in sorted(supported_args.items()):
                    print '{0:<20}{1}'.format(i, j)
                sys.exit(1)

    @staticmethod
    def print_element(element):
        for el in element:
            print el,

    def printer(self):
        if (self.pattern and self.message
            and 'Not supported' not in self.pattern
            and 'Not supported' not in self.message
            and 'Not implemented' not in self.pattern
            and 'Not implemented' not in self.message):
            sys.stdout = open('ok.txt', 'a+')
            ET.dump(self.rule)
            sys.stdout = sys.__stdout__

            RuleView.sys_argv(self)
            '''
            if not did_you_mean:
                print (self.rule.get('id') if self.rule.get('id')
                       else self.rule.getparent().get('id'))
            '''
            if self.key_to_print:
                if self.key_to_print in self.prefix:
                    if self.key_to_print in self.supported_keys:
                        sys.stdout = sys.__stdout__
                else:
                    sys.stdout = codecs.open(os.devnull, encoding='utf-8',
                                                                    mode='w')

            global number
            global amount
            number += 1
###############
            ''' API's "marker" difference handler'''
            if ('marker' in [el.tag
                        for el in self.rule.find('pattern').getchildren()]):

                ''' regexp in marker => suggestion_counters +=1'''

                for string in self.message:
                    if '\\' in string:
                        try:
                            parts = string.split('\\')
                            rest = []
                            for el in parts[1:]:
                                el = str(int(el[0]) + 1) + el[1:]
                                rest.append(el)
                            result = '\\'.join([parts[0], ] + rest)
                            self.message[self.message.index(string)] = result
                        except ValueError:
                            pass
            self.message.append('\n')
###############
            RuleView.print_element(self.pattern)
###############
            ''' API's '<- option[...] ->' handler'''
            print '<- option("%s")' % self.category.get('name'),
            try:
                ''' marker in pattern => -%d> in result'''
                marker = [el.tag for el
            in self.rule.find('pattern').getchildren()].index('marker') + 1
                if not 'skip' in self.keys:
                    print '-%d>' % marker,
                else:
                    print '-%d>' % (marker + 1),
            except ValueError:
                print '->',
###############
            RuleView.print_element(self.message)
############### For debug purposes only.
        elif ('Not implemented' in self.pattern
              or 'Not implemented' in self.message
              and 'Not supported' not in self.pattern
              and 'Not supported' not in self.message):
            global to_implement
            to_implement += 1
#            print ET.dump(self.rule)
            sys.stdout = open('notimplemented.txt', 'a+')
            ET.dump(self.rule)
            sys.stdout = sys.__stdout__
        elif ('Not supported' in self.pattern
              or 'Not supported' in self.message
              and 'Not implemented' not in self.pattern
              and 'Not implemented' not in self.message):
            global no_support
            no_support += 1
            sys.stdout = open('unsupported.txt', 'a+')
            ET.dump(self.rule)
            sys.stdout = sys.__stdout__


class RuleController(object):
    '''flow control'''
    def __init__(self):
        try:
            os.remove(os.path.join(os.getcwd(), 'unsupported.txt'))
        except OSError:
            pass
        try:
            os.remove(os.path.join(os.getcwd(), 'notimplemented.txt'))
        except OSError:
            pass
        try:
            os.remove(os.path.join(os.getcwd(), 'rules.txt'))
        except OSError:
            pass

        self.categories = ET.iterparse("grammar.xml", events=('end',),
                                       tag='category')

    def process(self):
        global number
        global amount
        for _event, category in self.categories:
            for rule in category.iter('rule'):
                amount += 1
                RuleView(rule, category).printer()
        if not did_you_mean:
            percent = float(number) / amount * 100
            print '\n%.3f %s of rules covered (%s/%s)' % (percent, chr(37),
                                                         number, amount)
            global to_implement
            print '%s rules left to cover' % to_implement
            global no_support
            print '%s unsupported rules' % no_support

RuleController().process()
